# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the utilities."""
import asyncio
from collections import defaultdict
from contextlib import asynccontextmanager
from functools import wraps
from typing import Any
from typing import AsyncGenerator
from typing import Awaitable
from typing import Callable
from typing import DefaultDict
from typing import Hashable
from typing import TypeVar

import anyio
import structlog

logger = structlog.get_logger()


CallbackType = Callable[..., Awaitable]
T = TypeVar("T")


def function_to_name(function: Callable) -> str:
    """Get a uniquely qualified name for a given function."""
    return function.__name__


def handle_exclusively(key: Callable[..., Hashable]) -> Callable:
    """Avoids race conditions in handlers by ensuring exclusivity for arguments.

    This decorator is used to ensure that the "same" message cannot be handled by a
    message handler more than once at the same time. Here, the "same" message refers
    to a message with arguments which would make the handling functionally identical,
    e.g. a MO message with identical `uuid` and `object_uuid`.

    Examples:
        Simple usage::

            @handle_exclusively(key=lambda x, y, z: x, y)
            async def f(x, y, z):
                pass

            await f(x=1, y=2, z=8)
            await f(x=3, y=4, z=8)  # accepted
            await f(x=1, y=2, z=9)  # blocked

    Args:
        key: A custom key function which should return the arguments considered when
         determining exclusivity.

    Returns: Wrapper.
    """
    locks: DefaultDict[Hashable, anyio.Lock] = defaultdict(anyio.Lock)

    @asynccontextmanager
    async def named_lock(key: Hashable) -> AsyncGenerator[None, None]:
        lock = locks[key]
        async with lock:
            try:
                yield
            finally:
                # Garbage collect lock if no others are waiting to acquire.
                # This MUST be done before we release the lock.
                if not lock.statistics().tasks_waiting:
                    del locks[key]

    def wrapper(coro: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
        @wraps(coro)
        async def wrapped(*args: Any, **kwargs: Any) -> T:
            async with named_lock(key=key(*args, **kwargs)):
                return await coro(*args, **kwargs)

        return wrapped

    return wrapper


@asynccontextmanager
async def sleep_on_error(delay: int = 30) -> AsyncGenerator:
    """Async context manager that delays returning on errors.

    This is used to prevent race-conditions on writes to MO/LoRa, when the upload times
    out initially but is completed by the backend afterwards. The sleep ensures that
    the AMQP message is not retried immediately, causing the handler to act on
    information which could become stale by the queued write. This happens because the
    backend does not implement fairness of requests, such that read operations can
    return soon-to-be stale data while a write operation is queued on another thread.

    Specifically, duplicate objects would be created when a write operation failed to
    complete within the timeout (but would be completed later), and the handler, during
    retry, read an outdated list of existing objects, and thus dispatched another
    (duplicate) write operation.

    See: https://redmine.magenta-aps.dk/issues/51949#note-23.
    """
    try:
        yield
    except Exception:  # pylint: disable=broad-except
        await asyncio.sleep(delay)
        raise


class RejectMessage(Exception):
    """Raise to reject a message, turning it into a dead letter."""


class RequeueMessage(Exception):
    """Raise to requeue a message for later redelivery."""



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
# pylint: disable=too-few-public-methods
"""This module contains the MO specific AMQP models."""
from datetime import datetime
from enum import auto
from enum import Enum
from enum import unique
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import cast
from typing import List
from typing import overload
from typing import Tuple
from uuid import UUID

from more_itertools import one
from pydantic import BaseModel
from pydantic.dataclasses import dataclass

from ..metrics import exception_mo_routing_counter


class AutoLowerNameEnum(Enum):
    """Enum subclass which lets `auto` assign the enum key as value.

    From: https://docs.python.org/3/library/enum.html#using-automatic-values

    Example:
        ```
        class Color(Enum):
            RED = auto()
            BLUE = auto()

        print(list(Color))
        ```
        Yields:
        ```
        [<Color.RED: 1>, <Color.BLUE: 2>]
        ```
        Whereas:
        ```
        class Color(str, AutoLowerNameEnum):
            RED = auto()
            BLUE = auto()

        print(list(Color))
        ```
        Yields:
        ```
        [<Color.RED: 'red'>, <Color.BLUE: 'blue'>]
        ```
    """

    @staticmethod
    def _generate_next_value_(
        name: Any, start: int, count: int, last_values: List[Any]
    ) -> str:
        # The next `auto` value is simply the key name
        return cast(str, name).lower()


@unique
class ServiceType(str, AutoLowerNameEnum):
    """The MO service types.

    Describes the root object for the object the operation was executed on.
    """

    EMPLOYEE: str = cast(str, auto())
    ORG_UNIT: str = cast(str, auto())
    WILDCARD: str = "*"


@unique
class ObjectType(str, AutoLowerNameEnum):
    """The MO object types.

    Describes the object the operation was executed on.
    """

    ADDRESS: str = cast(str, auto())
    ASSOCIATION: str = cast(str, auto())
    EMPLOYEE: str = cast(str, auto())
    ENGAGEMENT: str = cast(str, auto())
    IT: str = cast(str, auto())
    KLE: str = cast(str, auto())
    LEAVE: str = cast(str, auto())
    MANAGER: str = cast(str, auto())
    OWNER: str = cast(str, auto())
    ORG_UNIT: str = cast(str, auto())
    RELATED_UNIT: str = cast(str, auto())
    ROLE: str = cast(str, auto())
    WILDCARD: str = "*"


@unique
class RequestType(str, AutoLowerNameEnum):
    """The MO request types.

    Describes the type of operation which was executed.

    Args:
        CREATE: Something was created.
        EDIT: Something was edited.
        TERMINATE: Something was terminated.
        REFRESH: Something was requested to be refreshed.
        WILDCARD: Something happened.
    """

    CREATE: str = cast(str, auto())
    EDIT: str = cast(str, auto())
    TERMINATE: str = cast(str, auto())
    REFRESH: str = cast(str, auto())
    WILDCARD: str = "*"


class PayloadType(BaseModel):
    """MO AMQP message format.

    Args:
        uuid: The UUID of the org-unit or employee.
        object_uuid: The UUID of the object itself.
        time: Time when MO emitted the message.
    """

    uuid: UUID
    object_uuid: UUID
    time: datetime


@dataclass
class MORoutingKey:
    """MO Routing Key format.

    Args:
        service_type: The service type component.
        object_type: The object type to component.
        request_type: The request type to component.
    """

    service_type: ServiceType
    object_type: ObjectType
    request_type: RequestType

    @classmethod
    def from_tuple(
        cls, routing_key_tuple: Tuple[ServiceType, ObjectType, RequestType]
    ) -> "MORoutingKey":
        """Convert the given routing_key_tuple to a MORoutingKey.

        Args:
            routing_key_tuple: The routing tuple to convert.

        Returns:
            The equivalent MORoutingKey.
        """
        return cls(*routing_key_tuple)

    @classmethod
    def from_routing_key(cls, routing_key: str) -> "MORoutingKey":
        """Convert the given routing_key to a MORoutingKey.

        Raises:
            ValueError: Raised if the routing_key does not contain 2 periods.
            ValueError: Raised if the routing_key components cannot be parsed.

        Args:
            routing_key: The routing key to convert.

        Returns:
            The equivalent MORoutingKey.
        """
        with exception_mo_routing_counter.labels(routing_key).count_exceptions():
            parts = routing_key.split(".")
            if len(parts) != 3:
                raise ValueError("Expected a three tuple!")
            service_str, object_str, request_str = parts
            return cls(
                ServiceType(service_str),
                ObjectType(object_str),
                RequestType(request_str),
            )

    @overload
    @classmethod
    def build(
        cls,
        service_type: ServiceType,
        object_type: ObjectType,
        request_type: RequestType,
    ) -> "MORoutingKey":  # pragma: no cover
        ...

    @overload
    @classmethod
    def build(
        cls, routing_key_tuple: Tuple[ServiceType, ObjectType, RequestType]
    ) -> "MORoutingKey":  # pragma: no cover
        ...

    @overload
    @classmethod
    def build(cls, routing_key: str) -> "MORoutingKey":  # pragma: no cover
        ...

    @overload
    @classmethod
    def build(
        cls, mo_routing_key: "MORoutingKey"
    ) -> "MORoutingKey":  # pragma: no cover
        ...

    @classmethod
    def build(cls, *args: Any, **kwargs: Any) -> "MORoutingKey":
        """Attempt to construct a MORoutingKey.

        Returns:
            The constructed MORoutingKey.
        """
        num_arguments = len(args) + len(kwargs)

        if num_arguments == 3:
            return cls(*args, **kwargs)

        assert num_arguments == 1
        if args:
            arg = one(args)
        else:
            arg = kwargs[one(kwargs)]

        if isinstance(arg, MORoutingKey):
            return arg
        if isinstance(arg, tuple):
            return cls.from_tuple(
                cast(Tuple[ServiceType, ObjectType, RequestType], arg)
            )

        assert isinstance(arg, str)
        return cls.from_routing_key(arg)

    def __str__(self) -> str:
        """Convert this object to a routing_key.

        Args:
            self: The object to be converted.

        Returns:
            The equivalent routing key.
        """
        return ".".join(
            [self.service_type.value, self.object_type.value, self.request_type.value]
        )


MOCallbackType = Callable[..., Awaitable]



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
# flake8: noqa
"""This module defines the public interface of the RAMQP package."""
from .amqp import AMQPSystem
from .amqp import PublishMixin
from .amqp import Router



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
# pylint: disable=too-few-public-methods
"""This module contains the all pydantic BaseSetting(s)."""
from typing import Any
from typing import Dict
from typing import Optional

from pydantic import AmqpDsn
from pydantic import BaseSettings
from pydantic import parse_obj_as
from pydantic import root_validator
from pydantic import SecretStr


class ConnectionSettings(BaseSettings):
    """Settings for the AMQP connection.

    These settings can either be provided via environmental variables or directly to
    the AMQPSystem's init.
    """

    # The AMQP connection url, including credentials and vhost.
    url: AmqpDsn

    # Alternatively, AMQP connection settings can be set individually.
    scheme: Optional[str]
    host: Optional[str]
    user: Optional[str]
    password: Optional[SecretStr]
    port: Optional[int]
    vhost: Optional[str]

    # The AMQP Exchange we are binding queue messages from. Can be overridden
    # for integration specific exchanges.
    exchange: str = "os2mo"

    # Program specific queue name prefix, should be globally unique, but
    # consistent across program restarts. The program name is a good candidate.
    queue_prefix: Optional[str]

    # Maximum number of messages to fetch and handle in parallel.
    prefetch_count: int = 10

    @root_validator(pre=True)
    def url_from_fields(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Construct AMQP URI from individual fields if one is not set explicitly"""
        # Use full AMQP connection URI if set
        if values.get("url"):
            return values

        # Otherwise, construct one from the individual fields.
        # Pre-validators, which we need to use to have 'url' non-optional,
        # do not get defaults from untouched fields, so default values are
        # defined here instead of above.
        uri = "{scheme}://{user}:{password}@{host}:{port}".format(
            scheme=values.get("scheme", "amqp"),
            host=values.get("host", "msg_broker"),
            user=values.get("user", "guest"),
            password=values.get("password", "guest"),
            port=values.get("port", 5672),
        )
        if vhost := values.get("vhost"):
            uri = f"{uri}/{vhost}"
        values["url"] = parse_obj_as(AmqpDsn, uri)
        return values

    # pylint: disable=missing-class-docstring
    class Config:
        env_prefix = "amqp_"



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
# pylint: disable=too-few-public-methods
"""This module contains the MO specific AMQPSystem."""
from functools import wraps
from typing import Any
from typing import Callable
from typing import Dict
from typing import overload
from typing import Tuple

from aio_pika import IncomingMessage
from fastapi.encoders import jsonable_encoder
from pydantic import parse_raw_as

from ..abstract import AbstractAMQPSystem
from ..abstract import AbstractPublishMixin
from ..abstract import AbstractRouter
from ..mo.models import MOCallbackType
from ..mo.models import MORoutingKey
from ..mo.models import ObjectType
from ..mo.models import PayloadType
from ..mo.models import RequestType
from ..mo.models import ServiceType
from ..utils import CallbackType
from ..utils import handle_exclusively


class MORouter(AbstractRouter):
    """MO specific Router.

    Has specifically tailored `register` methods, which utilize the MO AMQP
    routing-key structure and payload format.
    """

    def __init__(self) -> None:
        super().__init__()
        self._adapter_map: Dict[MOCallbackType, CallbackType] = {}

    def _construct_adapter(self, adaptee: MOCallbackType) -> CallbackType:
        """Construct an adapter from a MOCallbackType to CallbackType.

        Note:
            Multiple calls with the same adaptee always return the same adapter.

        Args:
            adaptee: The callback function to be adapted.

        Returns:
            The adapted callback function calling the input adaptee on invocation.
        """

        # Early return the previously constructed adapter.
        # This is required so every function maps to one and only one adapter,
        # which is in itself required by start to ensure uniqueness of queues.
        if adaptee in self._adapter_map:
            return self._adapter_map[adaptee]

        def mo_payload_exclusivity_key(payload: PayloadType, **_: Any) -> tuple:
            return payload.uuid, payload.object_uuid

        exclusive_handler = handle_exclusively(key=mo_payload_exclusivity_key)

        @wraps(adaptee)
        async def adapter(message: IncomingMessage, context: dict) -> None:
            """Adapter function mapping MOCallbackType to CallbackType.

            Unpacks the provided message and converts it into a routing tuple and
            PayloadType before calling the captured callback.

            Args:
                message: incoming message to converted and passed to callback.
                context: Additional context from the AMQP system passed handlers.

            Returns:
                None
            """
            assert message.routing_key is not None
            mo_routing_key = MORoutingKey.from_routing_key(message.routing_key)
            payload = parse_raw_as(PayloadType, message.body)
            exclusive_adaptee = exclusive_handler(adaptee)
            await exclusive_adaptee(
                mo_routing_key=mo_routing_key, payload=payload, context=context
            )

        # Register our newly created adapter for early return.
        self._adapter_map[adaptee] = adapter
        return adapter

    @overload
    def register(
        self,
        service_type: ServiceType,
        object_type: ObjectType,
        request_type: RequestType,
    ) -> Callable:  # pragma: no cover
        ...

    @overload
    def register(
        self, routing_key_tuple: Tuple[ServiceType, ObjectType, RequestType]
    ) -> Callable:  # pragma: no cover
        ...

    @overload
    def register(self, routing_key: str) -> Callable:  # pragma: no cover
        ...

    @overload
    def register(self, mo_routing_key: MORoutingKey) -> Callable:  # pragma: no cover
        ...

    def register(self, *args: Any, **kwargs: Any) -> Callable:
        """Get a decorator for registering callbacks.

        Examples:
            ```
            address_create_decorator = morouter.register(
                ServiceType.EMPLOYEE,
                ObjectType.ADDRESS,
                RequestType.CREATE
            )
            @address_create_decorator
            def callback(
                service_type: ServiceType,
                object_type: ObjectType,
                request_type: RequestType,
                payload: PayloadType
            ) -> None:
                pass
            ```
            Or directly:
            ```
            @morouter.register(
                ServiceType.EMPLOYEE,
                ObjectType.ADDRESS,
                RequestType.CREATE
            )
            def callback(
                service_type: ServiceType,
                object_type: ObjectType,
                request_type: RequestType,
                payload: PayloadType
            ) -> None:
                pass
            ```

        Note that messages for each callback are always handled exclusively to avoid
        potential race conditions in the application. This means that multiple messages
        for the same MO PayloadType uuid/object_uuid cannot be handled concurrently.

        Args:
            service_type: The service type to bind messages for.
            object_type: The object type to bind messages for.
            request_type: The request type to bind messages for.

        Returns:
            A decorator for registering a function to receive callbacks.
        """
        mo_routing_key = MORoutingKey.build(*args, **kwargs)
        amqp_decorator = self._register(str(mo_routing_key))

        def decorator(function: MOCallbackType) -> MOCallbackType:
            amqp_decorator(self._construct_adapter(function))
            return function

        return decorator


class MOPublishMixin(AbstractPublishMixin):
    """MO specific PublishMixin.

    Has a specifically tailored `publish_message` method, which utilize the MO AMQP
    routing-key structure and payload format.
    """

    @overload
    async def publish_message(
        self,
        service_type: ServiceType,
        object_type: ObjectType,
        request_type: RequestType,
        payload: PayloadType,
    ) -> None:  # pragma: no cover
        ...

    @overload
    async def publish_message(
        self,
        routing_key_tuple: Tuple[ServiceType, ObjectType, RequestType],
        payload: PayloadType,
    ) -> None:  # pragma: no cover
        ...

    @overload
    async def publish_message(
        self, routing_key: str, payload: PayloadType
    ) -> None:  # pragma: no cover
        ...

    @overload
    async def publish_message(
        self, mo_routing_key: MORoutingKey, payload: PayloadType
    ) -> None:  # pragma: no cover
        ...

    async def publish_message(self, *args: Any, **kwargs: Any) -> None:
        """Publish a message to the given service.object.request tuple.

        Args:
            service_type: The service type to send the message to.
            object_type: The object type to send the message to.
            request_type: The request type to send the message to.
            payload: The message payload.

        Returns:
            None
        """
        if "payload" in kwargs:  # pragma: no cover
            payload = kwargs["payload"]
            del kwargs["payload"]
        else:
            payload = args[-1]
            args = args[:-1]

        routing_key = MORoutingKey.build(*args, **kwargs)
        payload_obj = jsonable_encoder(payload)
        await self._publish_message(str(routing_key), payload_obj)


class MOAMQPSystem(AbstractAMQPSystem[MORouter], MOPublishMixin):
    """MO specific AMQPSystem."""

    router_cls = MORouter



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ramqp', 'ramqp.mo']

package_data = \
{'': ['*']}

install_requires = \
['aio-pika>=8.3.0,<9.0.0',
 'anyio>=3.6.2,<4.0.0',
 'fastapi>=0.89.1,<0.90.0',
 'more-itertools>=9.0.0,<10.0.0',
 'prometheus-client>=0.16.0,<0.17.0',
 'pydantic>=1.10.4,<2.0.0',
 'ra-utils>=1.8.0,<2.0.0',
 'structlog>=22.3.0,<23.0.0']

setup_kwargs = {
    'name': 'ramqp',
    'version': '6.7.3',
    'description': 'Rammearkitektur AMQP library (aio_pika wrapper)',
    'long_description': '<!--\nSPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>\nSPDX-License-Identifier: MPL-2.0\n-->\n\n# Rammearkitektur AMQP\nRammearkitektur AMQP (RAMQP) is an opinionated library for AMQP.\n\nIt is implemented as a thin wrapper around `aio_pika`, with a generic and a MO\nspecific AMQPSystem abstract, the MO abstraction being implementing using a thin\nwrapper around the generic abstraction.\n\n## Usage\n\n### Generic\nReceiving:\n\n```python\nimport asyncio\nfrom typing import Any\n\nfrom ramqp import AMQPSystem\nfrom ramqp import Router\nfrom ramqp.config import ConnectionSettings\n\nrouter = Router()\n\n# Configure the callback function to receive messages for the two routing keys.\n# If an exception is thrown from the function, the message is not acknowledged.\n# Thus, it will be retried immediately.\n@router.register("my.routing.key")\n@router.register("my.other.routing.key")\nasync def callback_function(routing_key: str, **kwargs: Any) -> None:\n    pass\n\n\nasync def main() -> None:\n    settings = ConnectionSettings(queue_prefix="my-program")\n    async with AMQPSystem(settings=settings, router=router) as amqp_system:\n        await amqp_system.run_forever()\n\n\nasyncio.run(main())\n```\n**NOTE**: `**kwargs` is required in all handlers for forward compatibility: the\nframework can add new keywords in the future, and existing handlers should\naccept them without breaking, if they do not use them. \nIt can be named `**_`  to prevent the "unused variable" warning by linters.\n\n\nSending:\n```python\nfrom ramqp import AMQPSystem\n\nwith AMQPSystem() as amqp_system:\n    await amqp_system.publish_message("my.routing.key", {"key": "value"})\n```\n\n### MO AMQP\nReceiving:\n\n```python\nimport asyncio\nfrom typing import Any\n\nfrom ramqp.config import ConnectionSettings\nfrom ramqp.mo import MOAMQPSystem\nfrom ramqp.mo import MORouter\nfrom ramqp.mo.models import MORoutingKey\nfrom ramqp.mo.models import ObjectType\nfrom ramqp.mo.models import PayloadType\nfrom ramqp.mo.models import RequestType\nfrom ramqp.mo.models import ServiceType\n\nrouter = MORouter()\n\n\n# Configure the callback function to receive messages for the two routing keys.\n# If an exception is thrown from the function, the message is not acknowledged.\n# Thus, it will be retried immediately.\n@router.register(ServiceType.EMPLOYEE, ObjectType.ADDRESS, RequestType.EDIT)\n@router.register("employee.it.create")\nasync def callback_function(\n    mo_routing_key: MORoutingKey, payload: PayloadType, **kwargs: Any\n) -> None:\n    pass\n\n\nasync def main() -> None:\n    settings = ConnectionSettings(queue_prefix="my-program")\n    async with MOAMQPSystem(settings=settings, router=router) as amqp_system:\n        await amqp_system.run_forever()\n\n\nasyncio.run(main())\n```\n\nSending:\n\n```python\nfrom datetime import datetime\nfrom uuid import uuid4\n\nfrom ramqp.mo import MOAMQPSystem\nfrom ramqp.mo.models import ObjectType\nfrom ramqp.mo.models import PayloadType\nfrom ramqp.mo.models import RequestType\nfrom ramqp.mo.models import ServiceType\n\npayload = PayloadType(uuid=uuid4(), object_uuid=uuid4(), time=datetime.now())\n\nasync with MOAMQPSystem() as amqp_system:\n    await amqp_system.publish_message(\n        ServiceType.EMPLOYEE, ObjectType.ADDRESS, RequestType.EDIT, payload\n    )\n```\n\n### FastAPI and Additional Context\nThe `run_forever()` function is not very useful if you would like to serve an\nAPI at the same time. To solve this issue, the main entrypoint of the program is\nthe ASGI application, with the AMQP system tied to its lifespan. The example\nbelow also shows how additional context can be injected into the AMQP system,\nand subsequently retrieved in the handlers:\n\n```python\nfrom contextlib import asynccontextmanager\nfrom functools import partial\nfrom typing import Any\n\nfrom fastapi import APIRouter\nfrom fastapi import FastAPI\nfrom starlette.requests import Request\n\nfrom ramqp.config import ConnectionSettings\nfrom ramqp.mo import MOAMQPSystem\nfrom ramqp.mo import MORouter\n\namqp_router = MORouter()\nfastapi_router = APIRouter()\n\n\n@amqp_router.register("employee.it.create")\nasync def callback_function(context: dict, **kwargs: Any) -> None:\n    print(context["greeting"])  # Hello, world!\n\n\n@fastapi_router.get("/hello")\nasync def hello(request: Request) -> None:\n    # Passing the \'amqp_system\' in the context would allow FastAPI to actually publish\n    # this greeting to the whole world.\n    return request.app.state.context["greeting"]  # Hello, world!\n\n\ndef create_app() -> FastAPI:\n    context = {"greeting": "Hello, world!"}\n\n    app = FastAPI()\n    app.include_router(fastapi_router)\n    app.state.context = context  # share context with FastAPI\n\n    settings = ConnectionSettings(queue_prefix="my-program")\n    amqp_system = MOAMQPSystem(\n        settings=settings,\n        router=amqp_router,\n        context=context,  # share context with the AMQP system\n    )\n\n    app.router.lifespan_context = partial(lifespan, context, amqp_system)\n    return app\n\n\n@asynccontextmanager\nasync def lifespan(context: dict, amqp_system: MOAMQPSystem, app: FastAPI):\n    # Add to the context as required\n\n    async with amqp_system:\n        # Yield to keep the AMQP system open until the ASGI application is closed.\n        # Control will be returned to here when the ASGI application is shut down.\n        # Consider using AsyncExitStack in case of multiple context managers.\n        yield\n```\nSave the example as `example.py` and try it with\n`uvicorn --factory example:create_app`.\n\n\n### Metrics\nRAMQP exports a myraid of prometheus metrics via `prometheus/client_python`.\n\nThese can be exported using:\n```\nfrom prometheus_client import start_http_server\n\nstart_http_server(8000)\n```\nOr similar, see the promethues client library for details.\n\n\n## Development\n\n### Prerequisites\n- [Poetry](https://github.com/python-poetry/poetry)\n\n### Getting Started\n1. Clone the repository:\n```\ngit clone git@git.magenta.dk:rammearkitektur/ramqp.git\n```\n\n2. Install all dependencies:\n```\npoetry install\n```\n\n3. Set up pre-commit:\n```\npoetry run pre-commit install\n```\n\n### Running the tests\nYou use `poetry` and `pytest` to run the tests:\n\n`poetry run pytest`\n\nYou can also run specific files\n\n`poetry run pytest tests/<test_folder>/<test_file.py>`\n\nand even use filtering with `-k`\n\n`poetry run pytest -k "Manager"`\n\nYou can use the flags `-vx` where `v` prints the test & `x` makes the test stop if any tests fails (Verbose, X-fail)\n\n#### Running the integration tests\nTo run the integration tests, an AMQP instance must be available.\n\nIf an instance is already available, it can be used by configuring the `AMQP_URL`\nenvironmental variable. Alternatively a RabbitMQ can be started in docker, using:\n```\ndocker run -d -p 5672:5672 -p 15672:15672 rabbitmq:3-management\n```\n\n## Versioning\nThis project uses [Semantic Versioning](https://semver.org/) with the following strategy:\n- MAJOR: Incompatible changes to existing data models\n- MINOR: Backwards compatible updates to existing data models OR new models added\n- PATCH: Backwards compatible bug fixes\n\n\n## Authors\nMagenta ApS <https://magenta.dk>\n\n\n## License\nThis project uses: [MPL-2.0](MPL-2.0.txt)\n\nThis project uses [REUSE](https://reuse.software) for licensing.\nAll licenses can be found in the [LICENSES folder](LICENSES/) of the project.\n',
    'author': 'Magenta ApS',
    'author_email': 'info@magenta.dk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://magenta.dk/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# SPDX-FileCopyrightText: 2019-2020 Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the all prometheus metrics."""
import asyncio
from contextlib import contextmanager
from contextlib import ExitStack
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator

from aio_pika.abc import AbstractQueue
from aio_pika.abc import AbstractRobustChannel
from aio_pika.abc import AbstractRobustConnection
from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram

# --------------- #
# Various metrics #
# --------------- #

exception_decode_counter = Counter(
    "amqp_decode_exceptions",
    "Number of exceptions during body decoding",
    ["routing_key"],
)
exception_parse_counter = Counter(
    "amqp_parse_exceptions",
    "Number of exceptions during body parsing",
    ["routing_key"],
)
exception_mo_routing_counter = Counter(
    "amqp_mo_routing_exceptions",
    "Number of exceptions during mo routing key parsing",
    ["routing_key"],
)
routes_bound = Counter(
    "amqp_routes_bound", "Number of routing-keys bound to the queue", ["function"]
)
callbacks_registered = Counter(
    "amqp_callbacks_registered", "Number of callbacks registered", ["routing_key"]
)
event_counter = Counter("amqp_event", "Number of events made", ["event_key"])
event_last = Gauge("amqp_event_last", "Timestamp of last event", ["event_key"])


# --------------- #
# Receive metrics #
# --------------- #

receive_last_received = Gauge(
    "amqp_receive_last_received",
    "Timestamp of the last received message",
    ["routing_key", "function"],
)
receive_calls = Counter(
    "amqp_receive",
    "Number of calls made to receiver callbacks",
    ["routing_key", "function"],
)
receive_exceptions = Counter(
    "amqp_receive_exceptions",
    "Number of exceptions during receiver callbacks",
    ["routing_key", "function"],
)
receive_inprogress = Gauge(
    "amqp_receive_inprogress",
    "Number of receiver callbacks currently running",
    ["routing_key", "function"],
)
receive_time = Histogram(
    "amqp_receive_seconds",
    "Time spent running receiver callbacks",
    ["routing_key", "function"],
)


@contextmanager
def _handle_receive_metrics(
    routing_key: str, function_name: str
) -> Generator[None, None, None]:
    """Expose metrics about a callback.

    Args:
        routing_key: Routing key of the message.
        function_name: Name of the callback function.

    Returns:
        None
    """
    labels = (routing_key, function_name)

    receive_last_received.labels(*labels).set_to_current_time()
    receive_calls.labels(*labels).inc()
    with ExitStack() as stack:
        stack.enter_context(receive_exceptions.labels(*labels).count_exceptions())
        stack.enter_context(receive_inprogress.labels(*labels).track_inprogress())
        stack.enter_context(receive_time.labels(*labels).time())
        yield


# --------------- #
# Publish metrics #
# --------------- #

publish_last_published = Gauge(
    "amqp_publish_last_published",
    "Timestamp of the last published message",
    ["routing_key"],
)
publish_calls = Counter(
    "amqp_publish",
    "Number of messages published",
    ["routing_key"],
)
publish_exceptions = Counter(
    "amqp_publish_exceptions",
    "Number of exceptions during message publishing",
    ["routing_key"],
)
publish_inprogress = Gauge(
    "amqp_publish_inprogress",
    "Number of messages being published currently",
    ["routing_key"],
)
publish_time = Histogram(
    "amqp_publish_seconds",
    "Time spent publishing messages",
    ["routing_key"],
)


@contextmanager
def _handle_publish_metrics(routing_key: str) -> Generator[None, None, None]:
    """Expose metrics about a published message.

    Args:
        routing_key: Routing key of the message.

    Returns:
        None
    """
    labels = (routing_key,)

    publish_last_published.labels(*labels).set_to_current_time()
    publish_calls.labels(*labels).inc()
    with ExitStack() as stack:
        stack.enter_context(publish_exceptions.labels(*labels).count_exceptions())
        stack.enter_context(publish_inprogress.labels(*labels).track_inprogress())
        stack.enter_context(publish_time.labels(*labels).time())
        yield


# ---------------- #
# Periodic metrics #
# ---------------- #

last_periodic = Gauge("amqp_last_periodic", "Timestamp of the last periodic call")
last_loop_periodic = Gauge(
    "amqp_last_loop_periodic", "Timestamp (monotonic) of the last periodic call"
)
backlog_count = Gauge(
    "amqp_backlog",
    "Number of messages waiting for processing in the backlog",
    ["function"],
)


def _setup_periodic_metrics(queues: Dict[str, AbstractQueue]) -> asyncio.Task:
    """Setup a periodic job to update non-eventful metrics.

    Args:
        queues: The aio_pika queues to export periodic metrics for.

    Returns:
        asynchronous task for cancellation
    """
    # Setup periodic metrics
    async def periodic_metrics() -> None:
        loop = asyncio.get_running_loop()
        while True:
            last_periodic.set_to_current_time()
            last_loop_periodic.set(loop.time())
            for function_name, queue in queues.items():
                backlog_count.labels(function_name).set(
                    queue.declaration_result.message_count or 0
                )
            await asyncio.sleep(1)

    return asyncio.create_task(periodic_metrics())


# ------------------ #
# Connection metrics #
# ------------------ #


def event_callback_generator(event_key: str) -> Callable:
    """Generate event-keyed event callback function.

    Args:
        event_key: The key to label all metrics with.

    Returns:
        Callback function to update event metrics.
    """

    def event_callback(*_1: Any, **_2: Any) -> None:  # pragma: no cover
        """Reconnect callback to update reconnect metrics.

        Args:
            Unused arguments

        Returns:
            None
        """
        event_counter.labels(event_key).inc()
        event_last.labels(event_key).set_to_current_time()

    return event_callback


def _setup_connection_metrics(connection: AbstractRobustConnection) -> None:
    """Setup connection metrics collecting for reconnecting events.

    Args:
        connection: The connection install callbacks on.

    Returns:
        None
    """
    connection.reconnect_callbacks.add(
        event_callback_generator("RobustConnection.reconnect")
    )
    connection.close_callbacks.add(event_callback_generator("Connection.close"))


def _setup_channel_metrics(channel: AbstractRobustChannel) -> None:
    """Setup channel metrics collecting for reconnecting events.

    Args:
        connection: The channel install callbacks on.

    Returns:
        None
    """
    channel.reopen_callbacks.add(event_callback_generator("RobustChannel.reopen"))
    channel.close_callbacks.add(event_callback_generator("Channel.close"))
    channel.return_callbacks.add(event_callback_generator("Channel.return"))
